import json
from typing import List, Dict, Optional

from pydantic.main import BaseModel
from socketio import AsyncServer

from server.game.manager import ClientManager
from server.game.message import MessageToPlayer, MessageFromPlayer
from server.model.game import Game
from server.model.known import CARRE_ACE
from server.model.players import Player, Announce


class Metadata(BaseModel):
    ready: bool = False
    sid: str = ""
    last_announce: Optional[Announce] = None
    fresh_announce: bool = True


class LobbyManager(ClientManager):
    """ A ClientManager that handles a lobby, then orchestrates games.
    """

    def __init__(self, sio: AsyncServer):
        super().__init__()
        self.lobby: Dict[str, Player] = {}
        self.metadata: Dict[str, Metadata] = {}
        self.games: List[Game] = []
        self.sio = sio

    @property
    def nb_ready(self):
        return len(self.players_ready)

    @property
    def players_ready(self):
        return [self.lobby[k] for k, m in self.metadata.items() if m.ready]

    async def new_player(self, player: Player, sid: str) -> str:
        self.lobby[player.name] = player
        self.metadata[player.name] = Metadata(sid=sid)
        print(f"Added {player} to a lobby with {len(self.lobby)} players.")
        return f"Bienvenu, {player}! Il y a {len(self.lobby)} joueurs en ligne."

    def wants_to_play(self, player: Player):
        self.metadata[player.name].ready = True
        print(f"{player} ready to play! ({self.nb_ready} people ready)")

    async def announces(self, player: Player, announce: Announce):
        # FIXME: Call this message on incoming "ANNOUNCE"
        meta = self.metadata[player.name]
        meta.last_announce = announce
        meta.fresh_announce = True
        print(f"{player} ready to play! ({self.nb_ready} people ready)")

    def start_game(self, max_players=2):
        players = self.players_ready[:max_players]
        self.games.append(Game(players, manager=self))
        for p in players:
            self.metadata[p.name].ready = False

        print(f"Game started : {' vs '.join([p.name for p in players])}")

    def reset(self):
        players = len(self.players)
        games = len(self.games)
        msg = f"Resetting! sorry for the {players} players / {games} games..."
        print(msg)
        for p in self.players:
            m = self.metadata[p.name]
            self.send(p, MessageToPlayer.Reset)
            self.sio.disconnect(m.sid)
            self.lobby.clear()
            self.metadata.clear()
            self.games.clear()
        print(f"Reset done, {players} players, {games} games.")
        return msg

    def send(self, to: Player, message: MessageToPlayer, extra=None):
        data = {message: message.name}
        if extra:
            data["extra"] = extra
        self.sio.send(message)
        pass

    def handle_message(self, sid, data):
        message = None
        extras = {"players": [p.name for p in self.players]}

        print(f"Lobby| Received message from {sid}: {data}.")

        sender = self.which_player(sid)
        if sender:
            print(f"Lobby| Found sender: {sender.name}")

        for option in MessageFromPlayer:
            if option.value in data:

                if option == MessageFromPlayer.Waiting:
                    self.metadata[sender.name].ready = False

                    print(f"MSG|Player {sender.name} waiting, while {extras} ready.")

                    if len(self.players_ready) > 1:
                        message = MessageToPlayer.ReadyToStart

                elif option == MessageFromPlayer.Ready:
                    self.wants_to_play(player=sender)

                    print(f"MSG|Players ready: {extras}.")

                    message = MessageToPlayer.ReadyToStart \
                        if len(self.players_ready) > 1 else MessageToPlayer.Waiting
                    extras["playersReady"] = [p.name for p in self.players_ready]

                elif option == MessageFromPlayer.Bet:
                    # FIXME vraie annonce, pas juste carre d'as lol
                    self.announces(player=sender,
                                   announce=Announce(bet=CARRE_ACE))
                    game: Game = self.game_with(sender)
                    # TODO: connect with current game, return appropriate message
                elif option == MessageFromPlayer.Menteur:
                    self.announces(player=sender, announce=Announce())
                    # TODO: connect with current game, return appropriate message

        body = {"message": message.name if message else "none"}
        if extras:
            body["extras"] = extras
        return json.dumps(body)

    def which_player(self, sid) -> Player:
        sender: Optional[Player] = None
        for name, metadata in self.metadata.items():
            if metadata.sid == sid:
                sender = self.lobby[name]
        return sender

    def game_with(self, player: Player) -> Game:
        return [g for g in self.games if player in g.players][0]

    def send_waiting_for(self, player: Player):
        game = self.game_with(player)
        self.send(player, MessageToPlayer.YourTurn, extra={"bet": game.current_bet})
        for p in [p for p in game.players if p != player]:
            self.send(p, MessageToPlayer.Waiting, extra={"waitingFor": p.name})
